@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2024 Jonathan Perkin <jonathan@perkin.org.uk>
+ * Copyright (c) 2026 Jonathan Perkin <jonathan@perkin.org.uk>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -25,15 +25,33 @@ use std::error::Error;
 use std::fs;
 use std::fs::File;
 use std::io::{self, BufRead, BufReader};
+use std::net::ToSocketAddrs;
 use std::path::PathBuf;
 use std::process;
+use std::sync::LazyLock;
 use std::sync::atomic::{AtomicU64, Ordering};
-use std::time::Instant;
+use std::time::{Duration, Instant};
 use suppaftp::{FtpStream, types::FileType};
 use thiserror::Error;
 use url::Url;
 
 static FETCH_COUNTER: AtomicU64 = AtomicU64::new(0);
+static CONNECT_TIMEOUT: LazyLock<Duration> = LazyLock::new(|| {
+    Duration::from_secs(
+        env::var("MKTOOL_CONNECT_TIMEOUT")
+            .ok()
+            .and_then(|v| v.parse().ok())
+            .unwrap_or(15),
+    )
+});
+static READ_TIMEOUT: LazyLock<Duration> = LazyLock::new(|| {
+    Duration::from_secs(
+        env::var("MKTOOL_READ_TIMEOUT")
+            .ok()
+            .and_then(|v| v.parse().ok())
+            .unwrap_or(60),
+    )
+});
 
 #[derive(Args, Debug)]
 pub struct Fetch {
@@ -240,7 +258,18 @@ fn fetch_ftp(
 ) -> Result<u64, FetchError> {
     let host = url.host_str().ok_or(FetchError::NotFound)?;
     let path = url.path();
-    let mut ftp = FtpStream::connect((host, 21))?;
+    let port = url.port().unwrap_or(21);
+    let addr =
+        (host, port).to_socket_addrs()?.next().ok_or(FetchError::NotFound)?;
+    let stream = std::net::TcpStream::connect_timeout(&addr, *CONNECT_TIMEOUT)
+        .map_err(suppaftp::FtpError::ConnectionError)?;
+    stream
+        .set_read_timeout(Some(*READ_TIMEOUT))
+        .map_err(suppaftp::FtpError::ConnectionError)?;
+    stream
+        .set_write_timeout(Some(*READ_TIMEOUT))
+        .map_err(suppaftp::FtpError::ConnectionError)?;
+    let mut ftp = FtpStream::connect_with_stream(stream)?;
     ftp.login("anonymous", "anonymous")?;
     ftp.transfer_type(FileType::Binary)?;
     let mut ftpfile = ftp.retr_as_stream(path)?;
@@ -348,13 +377,7 @@ fn fetch_and_verify(
                                             "Verification failed for {url}: {e}"
                                         );
                                     });
-                                    if let Err(e) = fs::remove_file(&temp_name)
-                                    {
-                                        eprintln!(
-                                            "Failed to remove {}: {e}",
-                                            temp_name.display()
-                                        );
-                                    }
+                                    remove_temp(&temp_name);
                                     continue 'nextsite;
                                 }
                             }
@@ -366,12 +389,7 @@ fn fetch_and_verify(
                     progress.suspend(|| {
                         eprintln!("Unable to fetch {url}: {e}");
                     });
-                    if let Err(e) = fs::remove_file(&temp_name) {
-                        eprintln!(
-                            "Failed to remove {}: {e}",
-                            temp_name.display()
-                        );
-                    }
+                    remove_temp(&temp_name);
                     continue 'nextsite;
                 }
             }
@@ -404,7 +422,16 @@ fn fetch_and_verify(
                  * rename to final destination.
                  */
                 let tempfile = File::create(&temp_name)?;
-                body.copy_to(&mut progress.wrap_write(&tempfile))?;
+                if let Err(e) =
+                    body.copy_to(&mut progress.wrap_write(&tempfile))
+                {
+                    drop(tempfile);
+                    remove_temp(&temp_name);
+                    progress.suspend(|| {
+                        eprintln!("Unable to fetch {url}: {e}");
+                    });
+                    continue 'nextsite;
+                }
                 drop(tempfile);
                 if let Some(di) = distinfo {
                     if let Some(entry) = di.distfile(&file.filepath) {
@@ -415,12 +442,7 @@ fn fetch_and_verify(
                                         "Verification failed for {url}: {e}"
                                     );
                                 });
-                                if let Err(e) = fs::remove_file(&temp_name) {
-                                    eprintln!(
-                                        "Failed to remove {}: {e}",
-                                        temp_name.display()
-                                    );
-                                }
+                                remove_temp(&temp_name);
                                 continue 'nextsite;
                             }
                         }
@@ -450,11 +472,7 @@ fn fetch_and_verify(
             }
         }
     }
-    if let Err(e) = fs::remove_file(&temp_name) {
-        if e.kind() != io::ErrorKind::NotFound {
-            eprintln!("Failed to remove {}: {e}", temp_name.display());
-        }
-    }
+    remove_temp(&temp_name);
     Err(FetchError::NotFound)
 }
 
@@ -468,13 +486,22 @@ fn build_client() -> Result<Client, reqwest::Error> {
         .with_no_client_auth();
     Client::builder()
         .referer(false)
+        .connect_timeout(*CONNECT_TIMEOUT)
         .tls_backend_preconfigured(tls_config)
         .build()
 }
 
 #[cfg(not(feature = "webpki-roots"))]
 fn build_client() -> Result<Client, reqwest::Error> {
-    Client::builder().referer(false).build()
+    Client::builder().referer(false).connect_timeout(*CONNECT_TIMEOUT).build()
+}
+
+fn remove_temp(path: &PathBuf) {
+    if let Err(e) = fs::remove_file(path) {
+        if e.kind() != io::ErrorKind::NotFound {
+            eprintln!("Failed to remove {}: {e}", path.display());
+        }
+    }
 }
 
 /*
